-- ------------------------------------------------------------------------------ --
--                              TradeSkillMaster_Totals                           --
--             http://www.curse.com/addons/wow/tradeskillmaster_totals            --
--                                                                                --
--      An unofficial TradeSkillMaster Addon (http://tradeskillmaster.com)        --
-- ------------------------------------------------------------------------------ --

local TSM = select(2, ...)
local TSM = LibStub("AceAddon-3.0"):NewAddon(TSM,"TSM_TotalsBroker", "AceConsole-3.0","AceEvent-3.0")
local AceGUI = LibStub("AceGUI-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("TradeSkillMaster_TotalsBroker") -- loads the localization table
local LDB = LibStub( "LibDataBroker-1.1" )
local QT  = LibStub:GetLibrary( "LibQTip-1.0" )
local tooltip
local private = {}
local money 
local launcher
local realmName = GetCVar("realmName")
local faction = "none"

-- Constants
local TextYellow = "|cffffff00%s|r"
local TextWhite = "|cffffffff%s|r"
local TextGreen = "|cff00ff00%s|r"
local TextRed   = "|cffff0000%s|r"
local TextGold = "|cffffd70ag|r"
local TextSilver = "|cffc7c7cfs|r"
local TextCopper = "|cffeda55fc|r"
local IdInevitableSeal = 1129
local IdTemperedSeal = 994
local IdApexis = 823
local IdGarrisonResources = 824
local IdOil = 1101
local KeyGarrisonResources = "GarrisonResources"
local KeyOil = "Oil"
local KeyMoney = "Money"
local KeyApexis = "Apexis"
local KeyTemperedSeal = "TemperedSeal"
local KeyInevitableSeal = "InevitableSeal"
local GoldIcon = "|TInterface\\MoneyFrame\\UI-GoldIcon:0|t"
local SilverIcon = "|TInterface\\MoneyFrame\\UI-SilverIcon:0|t"
local CopperIcon = "|TInterface\\MoneyFrame\\UI-CopperIcon:0|t"
local icons = { 
				IdInevitableSeal = "|TInterface\\Icons\\achievement_battleground_templeofkotmogu_02_green",
				IdTemperedSeal = "|TInterface\\Icons\\ability_animusorbs",
				IdApexis = "|TInterface\\Icons\\inv_apexis_draenor",
				IdGarrisonResources = "|TInterface\\Icons\\inv_garrison_resource",
				IdOil = "|TInterface\\Icons\\garrison_oil"}

local entered = false
local initialValues = { }
local totalValues = { KeyGarrisonResources = 0, KeyOil = 0, KeyMoney = 0, KeyApexis = 0, KeyTemperedSeal = 0, KeyInevitableSeal = 0 }
local currentValues = { KeyGarrisonResources = 0, KeyOil = 0, KeyMoney = 0, KeyApexis = 0, KeyTemperedSeal = 0, KeyInevitableSeal = 0 }

-- ============================================================================
-- TSM Methods
-- ============================================================================
function TSM:OnInitialize()
	TSM.db = LibStub:GetLibrary("AceDB-3.0"):New("TradeSkillMaster_TotalsBrokerDB", dbDefaults, true)
	TSM.totals = LibStub:GetLibrary("AceDB-3.0"):New("TradeSkillMaster_TotalsDB", nil, true)
	
	for name, module in pairs(TSM.modules) do
		TSM[name] = module
	end

	TSM:RegisterEvent("PLAYER_ENTERING_WORLD")
	TSM:RegisterEvent("PLAYER_MONEY")
	TSM:RegisterEvent("CURRENCY_DISPLAY_UPDATE")
	
	TSM:RegisterModule()

	money = TSM.totals.global.calculatedTotals[realmName][faction]["total"]["Currencies"][KeyMoney]

	private:UpdateSummary()
end

function TSM:OnEnable()
end

function TSM:RegisterModule()
	-- TSM.icons = {
		-- { 
		-- side = "module", 
		-- desc = L["Totals"], 
		-- slashCommand = "totals", 
		-- callback = "Overview:Load", 
		-- icon = "Interface\\Icons\\INV_SCROLL_07" }
	-- }
	
	-- TSM.moduleOptions = {callback="Options:Load"}

	TSMAPI:NewModule(TSM)
end

-- ============================================================================
-- Event Handlers
-- ============================================================================
function TSM:PLAYER_ENTERING_WORLD()
	entered = true
	initialValues[KeyMoney] = GetMoney()
	private:UpdateSummary()
end

function TSM:CURRENCY_DISPLAY_UPDATE()
	private:UpdateSummary()
end

function TSM:PLAYER_MONEY()
	private:UpdateSummary()
end

-- ============================================================================
-- Private Helpers (should pull out into another class)
-- ============================================================================

function private:UpdateSummary()
	if not entered then
		return
	end
	money = TSM.totals.global.calculatedTotals[realmName][faction]["total"]["Currencies"][KeyMoney]

	local summary = private:MoneyToString(money).."  "
	
	summary = private:AddCurrencyTo(summary, IdGarrisonResources, KeyGarrisonResources)
	summary = private:AddCurrencyTo(summary, IdApexis, KeyApexis)
	summary = private:AddCurrencyTo(summary, IdOil, KeyOil)
	summary = private:AddCurrencyTo(summary, IdTemperedSeal, KeyTemperedSeal)
	summary = private:AddCurrencyTo(summary, IdInevitableSeal, KeyInevitableSeal)

	launcher.text = summary
end

function private:SessionCurrencyFor(id, currencyType, includeIcon, useChangedValue)
	useChangedValue = useChangedValue or false
	
	local colour = TextWhite
	local _, amount, icon = GetCurrencyInfo(id)
	local initialAmount

	currentValues[currencyType] = amount
	
	if not initialValues[currencyType] then
		initialValues[currencyType] = amount
		initialAmount = amount
	else
		initialAmount = initialValues[currencyType]
		
		if initialAmount ~= amount then
			colour = initialAmount < amount and TextGreen or TextRed 
		end
	end
	
	local prefix = ""
	
	if useChangedValue then
		amount = amount - initialValues[currencyType]
		if amount ~= 0 then
			prefix = amount > 0 and "+" or ""
			colour = amount > 0 and TextGreen or TextRed 
		end
	end
	
	if includeIcon then
		return "|T"..icon..":16|t" .. format(colour, private:FormatThousands(amount)) .. "  "
	else
		return format(colour, prefix .. private:FormatThousands(amount))
	end
end

function private:AddCurrencyTo(summary, id, currencyType)
	return summary .. private:SessionCurrencyFor(id, currencyType, true)
end

function private:ClassColourFor(name, class)
  local classColour = RAID_CLASS_COLORS[class] or {["r"] = 1, ["g"] = 1, ["b"] = 0}
  name = name or 0
  return format("|cff%02x%02x%02x%s|r", classColour["r"]*255, classColour["g"]*255, classColour["b"]*255, name)
end

local _G = _G
function RunSlashCmd(cmd)
  local slash, rest = cmd:match("^(%S+)%s*(.-)$")
  for name, func in pairs(SlashCmdList) do
     local i, slashCmd = 1
     repeat
        slashCmd, i = _G["SLASH_"..name..i], i + 1
        if slashCmd == slash then
           return true, func(rest)
        end
     until not slashCmd
  end
  -- Okay, so it's not a slash command. It may also be an emote.
  local i = 1
  while _G["EMOTE" .. i .. "_TOKEN"] do
     local j, cn = 2, _G["EMOTE" .. i .. "_CMD1"]
     while cn do
        if cn == slash then
           return true, DoEmote(_G["EMOTE" .. i .. "_TOKEN"], rest);
        end
        j, cn = j+1, _G["EMOTE" .. i .. "_CMD" .. j]
     end
     i = i + 1
  end
end

function private:FormatThousands(v)
	local prefix = ""
	if v < 0 then
		v = v * -1
		prefix = "-"
	end
	local s = string.format("%d", math.floor(v))
	local pos = string.len(s) % 3
	if pos == 0 then pos = 3 end
	
	return prefix .. string.sub(s, 1, pos)
		.. string.gsub(string.sub(s, pos+1), "(...)", ",%1")
end

function private:FormatNumber(num, pad, color)
	if num < 10 and pad then
		num = "0"..num
	end
	
	if color then
		return color..num.."|r"
	else
		return private:FormatThousands(num)
	end
end

function private:MoneyToString(money, ...)
	money = tonumber(money)
	if not money then return "---" end
	local color, isIcon, pad, trim, disabled = nil, nil, nil, nil, nil
	for i=1, select('#', ...) do
		local opt = select(i, ...)
		if type(opt) == "string" then
			if opt == "OPT_ICON" then -- use texture icons instead of letters for denominations
				isIcon = true
			elseif opt == "OPT_PAD" then -- left-pad all but the highest denomination with zeros (i.e. "1g 00s 02c" instead of "1g 0s 2c")
				pad = true
			elseif opt == "OPT_TRIM" then -- removes any 0 valued denominations (i.e. "1g" instead of "1g 0s 0c") - 0 will still be represented as "0c"
				trim = true
			elseif opt == "OPT_DISABLE" then -- removes color from denomination text - NOTE: this is not allowed if OPT_ICON is set
				disabled = true
			elseif strmatch(strlower(opt), "^|c[0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f][0-9a-f]$") then -- color the numbers
				color = opt
			else
				TSMAPI:Assert(false, "Invalid option: "..opt)
			end
		end
	end
	TSMAPI:Assert(not (isIcon and disabled), "Setting both OPT_ICON and OPT_DISABLE is not allowed")
	
	local isNegative = money < 0
	money = abs(money)
	local gold = floor(money / COPPER_PER_GOLD)
	local silver = floor((money % COPPER_PER_GOLD) / COPPER_PER_SILVER)
	local copper = floor(money % COPPER_PER_SILVER)
	local goldInidcator, silverInidcator, copperText = nil, nil, nil
	if isIcon then
		goldInidcator, silverInidcator, copperText = GoldIcon, SilverIcon, CopperIcon
	elseif disabled then
		goldInidcator, silverInidcator, copperText = "g", "s", "c"
	else
		goldInidcator, silverInidcator, copperText = TextGold, TextSilver, TextCopper
	end
	
	if money == 0 then
		return private:FormatNumber(0, false, color)..copperText
	end
	
	local text = nil
	local shouldPad = false
	if trim then
		wipe(private.textMoneyParts) -- avoid creating a new table every time
		-- add gold
		if gold > 0 then
			tinsert(private.textMoneyParts, private:FormatNumber(gold, false, color)..goldInidcator)
			shouldPad = pad
		else
		-- add silver
			tinsert(private.textMoneyParts, private:FormatNumber(0, shouldPad, color)..goldInidcator)
			shouldPad = pad
		end
		text = table.concat(private.textMoneyParts, " ")
	else
		if gold > 0 then
			text = private:FormatNumber(gold, false, color)..goldInidcator
		else
			text = private:FormatNumber(0, false, color)
		end
	end
	
	if isNegative then
		if color then
			return color.."-|r"..text
		else
			return "-"..text
		end
	else
		return text
	end
end

function spairs(t, order)
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

-- ============================================================================
-- Tooltip goodness
-- ============================================================================

launcher = LDB:NewDataObject("TSM_TotalsBroker", {
  type = "data source",
  text = private:MoneyToString(money),
  label = L["TSM_Totals"],
  icon = "Interface\\Minimap\\Tracking\\Auctioneer",

  OnEnter = function ( self )
    local columns = 7 -- title and gold

    tooltip = QT:Acquire("TotalsBrokerToolTip", columns)
    tooltip:SetScale(1)

    private:DrawTooltip()

    tooltip:SetAutoHideDelay(0.1, self)
    tooltip:EnableMouse()
    tooltip:SmartAnchorTo(self)
    tooltip:UpdateScrolling()
    tooltip:Show()
  end,
  
  OnClick = function(clickedframe, button)
	RunSlashCmd("/tsm totals")
  end,
} )

function private:AddRow(cells)
	local row = tooltip:AddLine(" ")
	local currentCell = 1
	
	for _, cell in pairs(cells) do
		local cellWidth = cell[3] or 1

		tooltip:SetCell(row, currentCell, cell[1], cell[2], cellWidth)
		currentCell = currentCell + cellWidth
	end
end 

function private:AddPeriod(name, period)
	local totals = TSM.totals.global.calculatedTotals[realmName][faction]
	local profit = totals[period]["Income"]["Sum"] - totals[period]["Expenses"]["Sum"]
	local colour = TextGreen

	private:AddRow({{format(TextYellow, name), "LEFT"}})
	private:AddRow({{format(TextWhite, L["Income"]), "LEFT"}, 
					{format(TextWhite, private:MoneyToString(totals[period]["Income"]["Sum"])), "RIGHT"}})
	private:AddRow({{format(TextWhite, L["Expenses"]), "LEFT"}, 
					{format(TextWhite, private:MoneyToString(totals[period]["Expenses"]["Sum"])), "RIGHT"}})
	private:AddRow({{format(TextWhite, L["Profit"]), "LEFT"}, 
					{format(profit < 0 and TextRed or TextGreen, private:MoneyToString(profit)), "RIGHT"}})

	tooltip:AddLine(" ")
end

function private:AddAllPeriods()
	local totals = TSM.totals.global.calculatedTotals[realmName][faction]
	local profitDaily = totals["daily"]["Income"]["Sum"] - totals["daily"]["Expenses"]["Sum"]
	local profitWeekly = totals["weekly"]["Income"]["Sum"] - totals["weekly"]["Expenses"]["Sum"]
	local profitMonthly = totals["monthly"]["Income"]["Sum"] - totals["monthly"]["Expenses"]["Sum"]
	local colour = TextGreen

	private:AddRow({{format(TextYellow, L["Today"]), "RIGHT", 2},
					{format(TextYellow, L["Last 7 Days"]), "RIGHT", 2},
					{format(TextYellow, L["Last 30 Days"]), "RIGHT", 3}
				  })
	private:AddRow({{format(TextWhite, L["Income"]), "RIGHT"}, 
					{format(TextWhite, private:MoneyToString(totals["daily"]["Income"]["Sum"])), "RIGHT"},
					{format(TextWhite, private:MoneyToString(totals["weekly"]["Income"]["Sum"])), "RIGHT", 2},
					{format(TextWhite, private:MoneyToString(totals["monthly"]["Income"]["Sum"])), "RIGHT", 3}
				  })
	private:AddRow({{format(TextWhite, L["Expenses"]), "RIGHT"}, 
					{format(TextWhite, private:MoneyToString(totals["daily"]["Expenses"]["Sum"])), "RIGHT"},
					{format(TextWhite, private:MoneyToString(totals["weekly"]["Expenses"]["Sum"])), "RIGHT", 2},
					{format(TextWhite, private:MoneyToString(totals["monthly"]["Expenses"]["Sum"])), "RIGHT", 3}
				  })
	private:AddRow({{format(TextWhite, L["Profit"]), "RIGHT"}, 
					{format(profitDaily < 0 and TextRed or TextGreen, private:MoneyToString(profitDaily)), "RIGHT"},
					{format(profitWeekly < 0 and TextRed or TextGreen, private:MoneyToString(profitWeekly)), "RIGHT", 2},
					{format(profitMonthly < 0 and TextRed or TextGreen, private:MoneyToString(profitMonthly)), "RIGHT", 3}
				  })

	tooltip:AddLine(" ")
end

function private:DrawTooltip()
	tooltip:Hide()
	tooltip:Clear()

	local currentLine = tooltip:AddHeader(" ")
	local totals = TSM.totals.global.calculatedTotals[realmName][faction]
	local profit = GetMoney() - initialValues[KeyMoney]
	
	tooltip:SetCell(currentLine, 1, format(TextWhite, L["TSM Totals"]), "CENTER", tooltip:GetColumnCount())
	tooltip:AddLine(" ")
	
	private:AddRow({{"", "CENTER"},
					{"|TInterface\\Minimap\\Tracking\\Auctioneer:18|t", "RIGHT"},
					{icons.IdGarrisonResources .. ":18|t", "RIGHT"},
					{icons.IdApexis .. ":18|t", "RIGHT"},
					{icons.IdOil .. ":18|t", "RIGHT"},
					{icons.IdTemperedSeal .. ":18|t", "RIGHT"},
					{icons.IdInevitableSeal .. ":18|t", "RIGHT"}})
	
	local alts = {}
	
	totals = TSM.totals.global.historicalTotals[realmName][faction]
	
	totalValues[KeyGarrisonResources] = 0
	totalValues[KeyOil] = 0
	totalValues[KeyApexis] = 0
	totalValues[KeyTemperedSeal] = 0
	totalValues[KeyInevitableSeal] = 0

	for name, _ in pairs(totals["total"]) do
		local money = totals["total"][name]["Currencies"][KeyMoney] or 0
		local garrisonRes = totals["total"][name]["Currencies"][KeyGarrisonResources]  or 0
		local oil = totals["total"][name]["Currencies"][KeyOil] or 0
		local apexis = totals["total"][name]["Currencies"][KeyApexis] or 0
		local tempered = totals["total"][name]["Currencies"][KeyTemperedSeal] or 0
		local inevitable = totals["total"][name]["Currencies"][KeyInevitableSeal] or 0
	
		totalValues[KeyGarrisonResources] = totalValues[KeyGarrisonResources] + garrisonRes
		totalValues[KeyOil] = totalValues[KeyOil] + oil
		totalValues[KeyApexis] = totalValues[KeyApexis] + apexis
		totalValues[KeyTemperedSeal] = totalValues[KeyTemperedSeal] + tempered
		totalValues[KeyInevitableSeal] = totalValues[KeyInevitableSeal] + inevitable
		
		if money ~= 0 or garrisonRes ~= 0 or oil ~= 0 or apexis ~= 0 or tempered ~= 0 or inevitable ~= 0 then
			if totals["total"][name]["Class"] then
				local class = totals["total"][name]["Class"]
				alts[name] = {money = money, garrisonRes = private:ClassColourFor(private:FormatThousands(garrisonRes), class), apexis = private:ClassColourFor(private:FormatThousands(apexis), class), oil = private:ClassColourFor(private:FormatThousands(oil), class),  tempered = private:ClassColourFor(private:FormatThousands(tempered), class), inevitable = private:ClassColourFor(private:FormatThousands(inevitable), class) } 
			else
				alts[name] = {money = money, garrisonRes = garrisonRes, apexis = apexis, oil = oil, tempered = tempered, inevitable = inevitable } 
			end
		end
	end

	for name, _ in spairs(alts) do
		local displayName = name
		
		if totals["total"][name]["Class"] then
			displayName = private:ClassColourFor(name, totals["total"][name]["Class"])
		end
		
		private:AddRow({{displayName, "RIGHT"},
						{private:MoneyToString(alts[name].money), "RIGHT"},
						{alts[name].garrisonRes, "RIGHT"},
						{alts[name].apexis, "RIGHT"},
						{alts[name].oil, "RIGHT"},
						{alts[name].tempered, "RIGHT"},
						{alts[name].inevitable, "RIGHT"}})
	end

	tooltip:AddSeparator()
	
	private:AddRow({{format(TextYellow, L["Totals"]), "RIGHT"},
					{private:MoneyToString(TSM.totals.global.calculatedTotals[realmName][faction]["total"]["Currencies"][KeyMoney]), "RIGHT"},
					{private:FormatThousands(totalValues[KeyGarrisonResources]), "RIGHT"},
					{private:FormatThousands(totalValues[KeyApexis]), "RIGHT"},
					{private:FormatThousands(totalValues[KeyOil]), "RIGHT"},
					{private:FormatThousands(totalValues[KeyTemperedSeal]), "RIGHT"},
					{private:FormatThousands(totalValues[KeyInevitableSeal]), "RIGHT"}
					})

	tooltip:AddSeparator()

	local colour = TextWhite
	
	if profit ~= 0 then
		colour = profit < 0 and TextRed or TextGreen
	end
	
	private:AddRow({{format(TextYellow, L["This Session"]), "RIGHT"},
					{format(colour, private:MoneyToString(profit)), "RIGHT"},
					{private:SessionCurrencyFor(IdGarrisonResources, KeyGarrisonResources, false, true), "RIGHT"},
					{private:SessionCurrencyFor(IdApexis, KeyApexis, false, true), "RIGHT"},
					{private:SessionCurrencyFor(IdOil, KeyOil, false, true), "RIGHT"},
					{private:SessionCurrencyFor(IdTemperedSeal, KeyTemperedSeal, false, true), "RIGHT"},
					{private:SessionCurrencyFor(IdInevitableSeal, KeyInevitableSeal, false, true), "RIGHT"}
					})

	tooltip:AddLine(" ")

	private:AddAllPeriods()
	-- private:AddPeriod(L["Today"], "daily")
	-- private:AddPeriod(L["Last 7 Days"], "weekly")
	-- private:AddPeriod(L["Last 30 Days"], "monthly")
end